/**
 * @fileoverview This file contains the server-side implementation of the
 * 'Fridge notes' Opera service in MVC pattern.
 *
 * It is intended to work on Opera with Unite feature enabled; uses opera.io,
 * opear.webserver, widget object and some JavaScript1.6 features.
 *
 * @requires server_utilities.js
 *
 * @author iYondaime (Tomasz Kurowski, tkurowski@opera.com)
 * @version 0.5
 */

// {%TRANSLATION%}
var strings = {
    // The two following messages appear or non-js clients
    READONLY : "Sorry, adding notes is not allowed right now.",
    CAPACITY : "The Fridge is full so your note could not be added.",
    
    // Service's data gets malformed only if the service owner
    // breaks it with his own hacked code. I'm not sure it should be translated, etc.
    DATA_MALFORMED : "Service data malformed."
};

function debug() {
    //opera.postError(arguments);
};
 
var modelWall = new WallModel();
var viewWall = new WallView();
var modelSession = new SessionModel();

var activityStream = new ActivityStream();

var controllerWall = new WallController(modelWall, viewWall, modelSession);

controllerWall.start();

/**
 * Constructs a new SessionModel object. Inteded as a Singleton.
 *
 * @class This is a class used for basic session handling. It sets
 * cookies to identify a user, for whom it keeps track of: user's name,
 * their email and post. When the service is restarted, the session
 * data resets.
 * @constructor
 * @return A new session model
 */
function SessionModel() {

    /*
     * Contains some configuration, of the {@link SessionModel}.
     *
     * @private
     * @type Object
     */
    var conf = {
        ADMIN_SID : 'admin',    // these are NOT used for authentication
        ADMIN_TOKEN : 'token',  // they're here just for completeness
        COOKIE_PERSISTANCE : 30  // in days
    };

    /**
     * Id of service's running session. Used for building
     * 'only-current-session-matching' users' sids.
     *
     * @private
     * @type int
     */
    var sessionInstanceId = (new Date()).valueOf();

    /*
     * Data of user's session. A hash array with sid as a key.
     *
     * @private
     * @type Object
     */
    var users = {};

    /**
     * Sets cookie header for a response.
     *
     * @private
     * @param {WebServerResponse} response the response for which
     * to set a header.
     * @param {String} key cookie's key (name).
     * @param {String} val a value to assign to the cookie.
     */
    function setCookie(response, key, val, days) {
        var expires = '';
        if (days) {
            var date = new Date();
            date.setTime(date.getTime() + (days * 24*60*60*1000));
            expires = "; expires="+date.toGMTString();
        }

        var cookie = ( escape(key) + "=" + escape(val) + expires + "\r\n\r\n");
        response.setResponseHeader("Set-Cookie", cookie);
    };

    /**
     * Parses a cookie from a request.
     *
     * This function is copied (almost letter-to-letter) from Yusef library
     * @author Yusef's library creators.
     * @private
     * @param {WebServerRequest} request the request to get a cookie from.
     * @param {String} key cookie's key (name).
     * @returns {String} a value assigned to the cookie, empty string if
     * no cookie found.
     */
    function getCookie(request, key) {
        // NOTE: it's disregarding this bit of the RFC:
        //       A server SHOULD also accept comma (,) as the separator
        //       between cookie-values for future compatibility.
        var regexp = new RegExp(key + '=("([^"]*)"|[^";]+)(;|$)');

        if("undefined" != typeof(request.headers.Cookie)) {
            for(var i = 0; i < request.headers.Cookie.length; ++i) {
                if(request.headers.Cookie[i].match(regexp)) {
                    return RegExp.$2 || RegExp.$1;
                }
            }
        }
        return '';
    };

    /**
     * Produces a unique user's Sessiond ID for current session.
     *
     * @private
     * @returns {String} a unique sid.
     */
    function makeUniqueId() {
        var chars = "abcdefghijklmnopqrstuvwxyz0123456789";
        var sid = '';

        do {
            for (var i = 0; i < 12; ++i)
                sid += chars[ Math.floor(Math.random() * chars.length)];
        } while ( (sid + sessionInstanceId) in users )

        return sid + sessionInstanceId;
    };

    /**
     * @see #getUserBySid
     * @private
     */
    function getUserBySid(sid) {
        return (sid in users) ? users[sid] : null;
    };

    /**
     * Assigns a post to a user (as to it's author).
     *
     * @param {String} postDate the post's date attribute.
     * @param {String} sid user's sid.
     */
    this.addPostToUser = function(postDate, sid) {
        var usr = getUserBySid(sid);
        if (usr) { usr.entries.push(postDate); }
    };

    /**
     * Checks if <i>user - entries</i> relation given in request is valid.
     *
     * @param {Object} requestData object as created by
     * {@link WallController#getRequestData} containing user's <b>sid</b> and
     * <b>remove</b> array of posts (posts' dates) the user's supposed to own.
     * @returns {boolean} true iff all posts in requestData.remove array
     * belong to the user.
     */
    this.userOwnsEntries = function(requestData) {
        var sid = requestData.sid || '';
        var entries = requestData.remove || [];

        var usr = getUserBySid(sid);
        if (!usr) return false;

        return entries.every(function(entry) {
            return !!~usr.entries.indexOf(entry)
        })
    };

    /**
     * Removes posts from user's session data.
     *
     * @param {Object} requestData object as created by
     * {@link WallController#getRequestData} containing user's <b>sid</b> and
     * <b>remove</b> array of posts (posts' dates) to remove from user's
     * session data.
     */
    this.removeUserEntries = function(requestData) {
        var sid = requestData.sid || '';
        var entries = requestData.remove || [];

        var usr = getUserBySid(sid);
        if (!usr) return;
        var remaining = [];

        remaining = usr.entries.filter( function(item) {
            return !~entries.indexOf(item);
        });
        usr.entries = remaining;
    };

    /**
     * Starts or updates a <i>client-server</i> session based on a request.
     *
     * @param {WebServerRequest} request user's request.
     * @param {Object} entry entry data (as entryData field of an object
     * returned by {@link WallController#getRequestData}). Optional.
     * @returns {String} a unique sid of the user.
     */
    this.setSession = function(request, entry, token) {
        var connection = request.connection;
        var sid = connection.isOwner
            ? conf.ADMIN_SID : getCookie(request, 'sid');

        var user = getUserBySid(sid);

        // add new user
        if (!user) {
            sid = connection.isOwner ? conf.ADMIN_SID : makeUniqueId();

            entry = entry || {};

            users[sid] = {
                author : entry.author || '',
                email : entry.email || '',
                entries : [],
                token : connection.isOwner ? conf.ADMIN_TOKEN : makeUniqueId()
            };
        }
        // update data if there's some
        else if (entry) {
            user.author = entry.author;
            user.email = entry.email;
        }

        // set/prolong user's session
        if (!connection.isOwner) {
            setCookie(
                connection.response, 'sid', sid, conf.COOKIE_PERSISTANCE
            );
        }

        return {sid : sid, token : users[sid].token};
    };

    /**
     * Returns session data of a user identified by sid.
     *
     * @param {String} sid user's sid.
     * @returns {String} user's data (email, name (as 'author') and entries)
     * of a user identified by sid or null if no such user exists.
     */
    this.getUserBySid = getUserBySid;
};




/**
 * Constructs a new WallController object. Inteded as a Singleton.
 *
 * @class This is Controller of the MVC based Fridge Notes service. It handles
 * requests and is responsible for communication. There is no communication
 * between models ({@link SessionModel}, {@link WallModel}) and view
 * ({@link WallView}) other than through WallController.
 * @param {WallModel} oModel an instance of {@link WallModel}.
 * @param {WallView} oView an instance of {@link WallView}.
 * @param {SessionModel} oSession an instance of {@link SessionModel}.
 * @constructor
 * @return A new wall controller
 */
function WallController(oModel, oView, oSession) {

    var conf = {
        MAX_POST_LENGTH : 128,
        MAX_INPUT_LENGTH : 64,
        DEFAULT_CAPACITY : 24
    };


    var pref = {
        get readonly() {
            var readonly = widget.preferenceForKey('readonly');
            return !(!readonly || readonly == 'no' || readonly == 'false');
        },

        set readonly(readonly) {
            var ro = !(!readonly || readonly == 'no' || readonly == 'false');
            widget.setPreferenceForKey(ro+'', 'readonly');
        },

        get capacity() {
            var capacity = widget.preferenceForKey('capacity');
            return capacity ? +capacity : conf.DEFAULT_CAPACITY;
        },

        set capacity(cap) {
            widget.setPreferenceForKey(cap, 'capacity');
        }
    };

    var errors = {
        unauthenticated : "Unauthenticated!",
        readonlyErr : strings.READONLY,
        capacityErr : strings.CAPACITY,
        nameTooLong : 'Please use a short nickname.',
        emailTooLong : 'Please use a shorter e-mail address.',
        postTooLong : 'Your post was too long.',
        unknownErr : "Unexpected error!",
        fishyRemoval : "Something fishy's going on: who want's to remove data?",
        fishyReadonly : "Something fishy's going on: "
            +"who want's to change 'readonly'?",
        fishyCapacity : "Something fishy's going on: "
            +"who want's to set capacity?",
        dataMalformed : strings.DATA_MALFORMED,
        parsingErr: "Error parsing XML Data from file."
    };

    var webserver = opera.io.webserver;
document.title = opera.io.webserver.currentServiceName;

    var xmlData = null;    // this will keep the xml DOM of the whole data
    var translator = new Translator({domain: 'service'});

    var notifier = new Notifier();

    /**
     * Construct a new LiveConnection object. Inteded as a Singleton.
     *
     * @class An inner class of {@link WallController} to handle live connections
     * with Ajax clients
     * @requires WallController uses some {@link WallController} private data.
     * @constructor
     * @return A new wall controller
     */
    function LiveConnections () {
        var connections = [];    // { sid, connection, revision };

        /**
         * Adds a new connection.
         *
         * @param {String} sid User's sid
         * @param {Object} conData object comprizing WebServerConnection
         * connection, and int revision fields.
        */
        this.add = function (conData) {
            connections.push(conData);
        };

        /**
         * Builds and sends response to all awaiting Ajax clients.
         *
         * @param {String} action action indicating how to build the response
         * @param {Object} actionData object with data specific to the action
         */
        this.sendResponse = function (action, actionData) {
            var responseText = '',
                htmlData = null;

            connections.forEach(function(con) {
                if (con.connection.closed) {return;}

                switch (action) {
                    case 'save':
                        responseText = buildResponse(
                            con.revision,
                            actionData,
                            {
                                isAdmin : con.connection.isOwner,
                                user : oSession.getUserBySid(con.sid)
                            });
                    break;
                    case 'remove':
                        con.connection.response.setResponseHeader(
                            'X-Remove',
                            Array.prototype.join.call(actionData,',')
                        );
                    break;
                    default:
                    break;
                }

                // send resposne
                setConfigHeaders(con.connection.response);
                if (responseText) {
                    //con.connection.response.write(responseText);
                    writeResponse( con.connection, responseText );
                }
                con.connection.response.close();
            });
            connections = [];
        };

    };

    var liveConnections = new LiveConnections();

    /**
     * Builds an object containing parsed and validated request's data & info.
     *
     * @param {WebServerRequest} request request to get the data from
     * @returns {Object} requestData object containg request data & info
     * and validation results
     * @private
     */
    function getRequestData(request) {

        var requestData = {
            entryData : null,     // !null if there's sth sent in request
            error : '',           // non-empty iff there's an error...
            remove : [],
            readonly : false,
            revision: 0,
            capacity: -1,
            sid : ''
        };

        //> national font won't work for request.bodyItems[]...
        //var bodyItems = request.bodyItems;
        var bodyItems = toBodyItems(request.body);

        // if ('Revision' in request.headers) won't work for IE!!!
        var aRevision;
        if (aRevision = request.getRequestHeader('X-Revision')) {
            requestData.revision = parseInt(aRevision[0]);
        }

        // build up requestData object
        if (
            ('author' in bodyItems)
            && ('email' in bodyItems)
            && ('post' in bodyItems)
        ) {
            requestData.entryData = {};

            var decoded = '';
            var keys = ['author','email','post'];
            for (var i=0, key; key = keys[i]; ++i) {
                decoded = decodeURIComponent(bodyItems[key][0]);
                if (key == 'post') {
                    if (decoded.length > conf.MAX_POST_LENGTH) {
                        requestData.error = _('Your note can only be {characters} characters.');
                        requestData.entryData = null;
                        break;
                    }
                    else if (decoded.trim() === '') {
                        requestData.error = _('Empty notes are not allowed.');
                        requestData.entryData = null;
                        break;
                    }
                }
                else {
                    decoded = decoded.trim().substr(0, conf.MAX_INPUT_LENGTH);
                }

                decoded = decoded.entityify();
                
                requestData.entryData[key] = key === 'post' 
                    ? decoded.replace(/\n/g,'<br/>').replace(/  /g,'<dbsp/>')
                    : decoded;

            }
        }
        else if ('remove' in bodyItems) {
            requestData.remove = bodyItems['remove'];
        }
        else if ('readonly' in bodyItems) {
            requestData.readonly = bodyItems['readonly'][0];
        }
        else if ('capacity' in bodyItems) {
            var capacity = bodyItems['capacity'][0];
            capacity = parseInt(capacity);
            if (isNaN(capacity) || capacity < 0) {
                requestData.error = _('Please type a number.');
            }
            else {
                requestData.capacity = capacity;
            }
        }

        requestData.token = 'token' in bodyItems ? bodyItems['token'][0] : '';
        var session = oSession.setSession(request, requestData.entryData, requestData.token);
        
        requestData.sid = session.sid;
        requestData.authenticated = request.connection.isOwner
            ? true : session.token == requestData.token;

        return requestData;
    };

    /**
     * Builds an object containing response's data & info.
     *
     * @param {int} revision client's revision. Used to determine if client
     * used ajax request and if they're up to date.
     * @param {XMLDocumentFragment} xmlEntry unless null, it contains xml data
     * that should be send to client
     * @param {Object} usrData object with info about the user, containing
     * isAdmin field and user field as obtained by {@link
     * SessionModel#getUserById}.
     * @returns {String} responseText string to be sent in response
     * @private
     */
    function buildResponse(revision, xmlEntry, usrData) {
        // usrData = { isAdmin : isOwner, user : oSession.getUserBySid() }
        var responseText = '';
        var upToDate = false;
        var htmlData = null;

        // AJAX and up-to-date?
        if (revision && revision >= oModel.getPrevRevision() ) {
            upToDate = true;

            if (xmlEntry) {
                htmlData = oView.xmlToHtml(xmlEntry, usrData);
            }
        }

        if (!upToDate) {
            // build from strach
            if (xmlData) {
                htmlData = oView.xmlToHtml(xmlData.documentElement, usrData);
            }
            else {
                debug(errors['parsingErr']);    // ref{xmlData}
                return '';
            }
        }

        if (!revision) {    // FRESH PAGE, htmlData != null
            // open template and append htmlData to shout box
            var htmlPage = oView.completePage(
                htmlData, usrData, {
                    readonly : pref.readonly,
                    capacity : pref.capacity,
                    entriesNumber : oModel.getEntriesNumber()
                }
            );

            responseText = '<!DOCTYPE html PUBLIC '
                +'"-//W3C//DTD XHTML 1.0 Strict//EN"'
                +' "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">';
            responseText += (new XMLSerializer).serializeToString(
                htmlPage.documentElement
            );
            
            var replacementHashMap =
            {
                servicePath :''+opera.io.webserver.currentServicePath,
                deviceName  :''+opera.io.webserver.deviceName,
                username    :''+opera.io.webserver.userName,
                serviceName :''+opera.io.webserver.currentServiceName,
                hostName    :''+opera.io.webserver.hostName
            }
            for( var k in replacementHashMap )
            {
                var v = replacementHashMap[k];
                if( typeof(v)==='string' )
                {
                    responseText = responseText.replace( new RegExp( '@@'+ k +'@@', 'g' ), v );
                }
            }
            
        }
        else if (htmlData) {
            responseText = (new XMLSerializer).serializeToString(htmlData);
        }

        return responseText;
    };

    /**
     * Outputs the string to <code>connection</code>.response and translates
     * it at the same time.
     *
     * @param {WebServerConnection} connection the currently handeld
     * connection.
     * @param {String} string String to be written to the output.
     * @private
     */
    function writeResponse( connection, string )
    {
        string = translateResponse( connection, string );
        connection.response.write( string );
    }

    /**
     * Translates the given string using the translation library and the Accept-Language HTTP header.
     *
     * @param {WebServerConnection} connection the currently handeld
     * connection.
     * @param {String} string The string to be translated.
     * @return the <code>string</code> translated.
     * @private
     */
    function translateResponse( connection, string )
    {        
        return translator.translate(string);
    }

    //function setupUnderscoreFunction( connection )
    //{
    //    var langs = (connection.request.getRequestHeader( 'Accept-Language' )||[])[0];
    //    translator.setLanguages( langs );
    //}
    
    window._ = translator.underscore;


    /**
     * Deals with a connection, when the request was empty.
     *
     * As the request is empty, this method closes only the
     * connection it directly handles. In particluar: it does not call {@link
     * LiveConnections#sendResponse}. It will also be used when error occurs
     * in 'non-empty' request, so as awaiting connections are not broken
     * unneccessarly.
     * @param {WebServerConnection} connection the currently handeld
     * connection.
     * @param {Object} requestData object returned by {@link
     * WallController#getRequestData}
     * @private
     */
    function onEmptyRequest(connection, requestData) {

        var response = connection.response;

        var responseText = buildResponse(
            requestData.revision,
            null,
            {
                isAdmin : connection.isOwner,
                user : oSession.getUserBySid(requestData.sid)
            }
        );

        if (!responseText) { return; }  // hold on the connection

        setConfigHeaders(response);
        //response.write(responseText);
        writeResponse( connection, responseText );
        response.close();
    };

    /**
     * Deals with a connection, when a post was sent with request.
     *
     * Attempts to save the post ({@link WallModel#savePost}). Assigns
     * the post to the user ({@link SessionModel#addPostToUser})
     * sends the response for awaiting connections ({@link
     * LiveConnections#sendResponse}).
     * @param {WebServerConnection} connection the currently handeld
     * connection.
     * @param {Object} requestData object returned by {@link
     * WallController#getRequestData}
     * @returns {String} containing error key, or <b>false</b> if
     * everything was all right.
     * @private
     */
    function onDataSent(connection, requestData) {
        if (!requestData.authenticated) {return 'unauthenticated';}
        
        var entry = requestData.entryData,
            sid = requestData.sid;

        var xmlEntry = null;

        if (pref.readonly && !connection.isOwner) {return 'readonlyErr';}

        if (
            oModel.getEntriesNumber() >= pref.capacity
            && !connection.isOwner
        ) { return 'capacityErr'; }

        // save sent data
        xmlEntry = oModel.savePost(entry);

        if (!xmlEntry) {return 'unexpectedErr';}

        oSession.addPostToUser(
            xmlEntry.documentElement.getAttribute('date'), sid
        );
        var fridgeLink =  'http://'+connection.request.headers.Host[0] + opera.io.webserver.currentServicePath;
        // TODO: if we have a direct link to the note, we should have posted this as well
        activityStream.publish(fridgeLink, null,
                               new ActivityObjectModel(ActivityObjectType.Note, (entry.author ? entry.author : _('Somebody')), fridgeLink,
                                                       { content : entry['post'] }
                                                      ), entry['post']);

        // push the data to all awaiting connections
        liveConnections.sendResponse('save', xmlEntry);

        if (!requestData.revision) {
            onEmptyRequest(connection, requestData);
        }

        // notify
        if (!connection.isOwner) {
            translator.setLanguages(navigator.userLanguage);
            notifier.notify(_('{name} put a note on your Fridge!', {'name': (entry.author ? entry.author : _('Somebody'))}));
            translator.setLanguagesFromHttp(connection);
        };

        return false;
    };

    /**
     * Deals with a connection, when request was about removing posts.
     *
     * @param {WebServerConnection} connection the currently handeld
     * connection.
     * @param {Object} requestData object returned by {@link
     * WallController#getRequestData}
     * @returns {String} containing error key, or <b>false</b> if
     * everything was all right.
     * @private
     */
    function onRemoveRequest(connection, requestData) {
        if (!requestData.authenticated) {return 'unauthenticated';}
        
        if (! oSession.userOwnsEntries(requestData) && ! connection.isOwner) {
            return 'fishyRemoval';
        }

        if (pref.readonly && !connection.isOwner) {return 'readonlyErr';}

        oModel.removeEntries(requestData.remove);
        xmlData = oModel.getXmlData();
        oSession.removeUserEntries(requestData);    // label{xmlData}

        liveConnections.sendResponse('remove', requestData.remove)

        if (!requestData.revision) {
            onEmptyRequest(connection, requestData);
        }

        // notify
        if (!connection.isOwner) {
            var usr = oSession.getUserBySid(requestData.sid);
            // usr should never be null here... but... just in case...
            usr = usr ? usr : {};
            notifier.notify(_('{name} deleted their note.', {name: (usr.author ? usr.author : 'Somebody')}));
        };

        return false;
    };

    /**
     * Deals with a connection, when request was about setting readonly.
     *
     * @param {WebServerConnection} connection the currently handeld
     * connection.
     * @param {Object} requestData object returned by {@link
     * WallController#getRequestData}
     * @returns {String} containing error key, or <b>false</b> if
     * everything was all right.
     * @private
     */
    function onReadonlyRequest(connection, requestData) {

        if (!connection.isOwner) {return 'fishyReadonly';}

        pref.readonly = requestData.readonly;

        liveConnections.sendResponse();

        if (!requestData.revision) {
            onEmptyRequest(connection, requestData);
        }

        return false;
    };

    /**
     * Deals with a connection, when request was about setting capacity.
     *
     * @param {WebServerConnection} connection the currently handeld
     * connection.
     * @param {Object} requestData object returned by {@link
     * WallController#getRequestData}
     * @returns {String} containing error key, or <b>false</b> if
     * everything was all right.
     * @private
     */
    function onCapacityRequest(connection, requestData) {
        if (!connection.isOwner) {return 'fishyCapacity';}

        pref.capacity = requestData.capacity;

        liveConnections.sendResponse();

        if (!requestData.revision) {
            onEmptyRequest(connection, requestData);
        }

        return false;
    };

    /**
     * Puts to the response headers info about model's configuration
     * (revision, readonly).
     *
     * @param {WebServerResponse} response, which headers are to be set
     */
    function setConfigHeaders(response) {

        response.setResponseHeader('X-Revision', oModel.getRevision());
        response.setResponseHeader('X-Prev-revision', oModel.getPrevRevision());

        response.setResponseHeader(
            'X-Readonly',
            pref.readonly || pref.capacity <= oModel.getEntriesNumber()
        );

        response.setResponseHeader('Cache-control', 'no-cache');
        response.setResponseHeader('Content-type', 'text/html; charset=utf-8');

        if (response.connection.isOwner) {
            response.setResponseHeader('X-Capacity', pref.capacity)
        }
    };

    /**
     * Closes a response after sending error message.
     *
     * If client's using ajax it puts the error message to the 'Error' header.
     * Otherwise it writes the message directly to the response.
     * @param {WebServerResponse} response response to be closed.
     * @param {String} errorMsg error message to be sent to the client.
     * @param {boolean} ajax whether client uses ajax.
     * @param {boolean} fatal set's 'Fatal-error' instead of 'Error' header,
     * which means that client is not to try to reconnect automatically.
     */
    function onError(response, errorMsg, ajax, fatal) {
        setConfigHeaders(response);
        if (ajax) {
            var headerKind = fatal ? 'X-Fatal-error' : 'X-Error';
            response.setResponseHeader(headerKind, errorMsg);
        }
        else {
            response.write(errorMsg);
        }
        response.close();
    };

    /**
     * Handles <i>_index</i> WebServerRequest.
     *
     * Answers all sort of requests the Fridge Notes service provides.
     * @param {WebServerRequestEvent} evt an event of a request being made.
     */
    function onIndexAction(evt) {

        var request = evt.connection.request,
            response = evt.connection.response;

        // build request object
        var requestData = getRequestData(request);

        //setupUnderscoreFunction(evt.connection);
        translator.setLanguagesFromHttp(evt.connection);

        // parse XML data from storage
        xmlData = oModel.getXmlData();    //> on NULL
        if (!xmlData) {
            onError(response, _(errors['dataMalformed']), requestData.revision, true);
            debug(errors['dataMalformed']);
            return;
        }

        // behave appropriately to content of requestData
        if (requestData.error) {
            onError(response, requestData.error, requestData.revision);
            return;
        }

        // if client's using AJAX, create push-poll connection
        if (requestData.revision) {
            liveConnections.add({
                sid : requestData.sid,
                connection : evt.connection,
                revision : requestData.revision
            });
        }
        // client sent some data?
        if (requestData.entryData) {
            var errKey = onDataSent(evt.connection, requestData);
            if (errKey) {
                onError(response, _(errors[errKey]), requestData.revision);
            }
        }
        else if (requestData.remove.length) {
            var errKey = onRemoveRequest(evt.connection, requestData);
            if (errKey == 'readonlyErr' || errKey == 'unauthenticated') {
                onError(response, _(errors[errKey]), requestData.revision);
            }
            else if (errKey) {
                onEmptyRequest(evt.connection, requestData);
                debug(errors[errKey]);
            };
        }
        else if (requestData.readonly) {
            var errKey = onReadonlyRequest(evt.connection, requestData);
            if (errKey) {
                onEmptyRequest(evt.connection, requestData);
                debug(errors[errKey]);
            };
        }
        else if (requestData.capacity >= 0) {   // or maybe > 0?
            var errKey = onCapacityRequest(evt.connection, requestData);
            if (errKey) {
                onError(response, _(errors[errKey]), requestData.revision);
            }
        }
        else {
            var errKey = onEmptyRequest(evt.connection, requestData);
        }

    };

    /**
     * Starts the service.
     */
    this.start = function() {
        var success = true;
        if ( !oModel.getXmlData() ) { success = oModel.createNewXml(); }

        if (success) {
            webserver.addEventListener('_request', function( event )
            {
                var path = event.connection.request.uri.split('?')[0];
                if( path == opera.io.webserver.currentServicePath )
                {
                    onIndexAction.apply( this, arguments );
                }
                else if( path == opera.io.webserver.currentServicePath + 'translation/client.js' )
                {
                    translator.sendClientScript( event.connection );
                }
                else if( !event.connection.closed )
                {
                    event.connection.response.closeAndRedispatch();
                }
            }, false);
            activityStream.listen();
        }
    };

};




/**
 * Constructs a new WallModel object. Inteded as a Singleton.
 *
 * @class This is Model of the MVC based Fridge Notes service. Provides
 * means of saving and manipulating service's data using opera.io.filesystem.
 * @constructor
 * @return A new wall model
 */
function WallModel() {

    var filesystem = opera.io.filesystem;
    var storage = filesystem.mountSystemDirectory('storage');
    var fileWall = storage.resolve('wall.xml');

    /*
    <?xml version="1.0" encoding="utf-8"?>
    <wall>
        <entry date="2009-03-23">
            <nick>itachi</nick>
            <email>tkurowski@opera.com</email>
            <post>My first post on the wall!</post>
        </entry>
        <!-- ... -->
    </wall>
    */
    var xmlWall;    //> = document.implementation.createDocument("","",null);
                    //> xmlWall.async = false;

    var prevRevision = 0,
        revision = 0;

    /**
     * Parses service data from XML file.
     *
     * @private
     * @param {File} file to be read
     * @returns {XMLDocument} XMLDocument object or null if an error occured
     * @private
     */
    function parseXmlFromFile(file) {

        if ( (typeof file) == 'undefined' || !file.exists ) { return null; }

        //> xmlDoc.load(file.path)    // why won't that work?!
        try {
            var strmWall = file.open(null, opera.io.filemode.READ);
            var sWall = '';
            while (!strmWall.eof) { sWall += strmWall.readLine(); }
            strmWall.close();
        }
        catch (e) { return null }

        var xmlDoc = (new DOMParser()).parseFromString(sWall, "text/xml");

        // check for errors
        var roottag = xmlDoc.documentElement;
        if (roottag.tagName != "wall") { return null; }

        return xmlDoc;
    };

    /**
     * Puts data to XML document.
     *
     * @private
     * @param {Object} oEntry object (JSON equivalent?) of xml wall entry (the
     * entryData field of what {@link WallController#getRequestData} returns).
     * @param {XMLDocumement} xmlDoc the document the data is put to
     * @param {String} sRevision the entrys date attribute and wall's revision
     * attribute value to be set
     * @returns {XMLDocumentFragment} xml fragment (the oEntry-to-xml
     * transformation result, which copy is appended to xmlDoc)
     */
    function addDataToXml(oEntry, xmlDoc, sRevision) {

        var aEntry = ['<?xml version="1.0"?><entry date="',sRevision,
            '"><author>',oEntry.author,'</author><email>',
            oEntry.email,'</email><post>',
            oEntry.post,'</post></entry>'];

        var sEntry = aEntry.join('');

        var xmlEntry = (new DOMParser()).parseFromString(sEntry, "text/xml");

        // add xmlEntry to xmlDoc
        var wall = xmlDoc.getElementsByTagName('wall')[0];

        wall.setAttribute('revision',sRevision);
        wall.appendChild(xmlEntry.documentElement.cloneNode(true));

        // on write success
        return xmlEntry;
    };

    /**
     * Saves XMLDocument to File.
     *
     * @private
     * @param {XMLDocument} xmlDoc document to be saved
     * @param {File} file the data is written to
     * @returns {Boolean} true on success, false otherwise
     */
    function saveXmlInFile(xmlDoc, file) {
        try {
            var strmWall = file.open(null, opera.io.filemode.WRITE);
            strmWall.write((new XMLSerializer()).serializeToString(xmlDoc));
            strmWall.close();
        }
        catch (e) {
            debug('The file could not be saved?!');
            return false
        }
        return true
    };

    /**
     * Writes empty XML wall structure to a file.
     *
     * @private
     * @param {File} file where to write the data
     * @param {String} sRevision revision attribute value
     * @returns {Boolean} true on success, false otherwise
     */
    function createNewXmlInFile(file, sRevision) {
        var sWall = '<?xml version="1.0"?><wall revision="'
            +sRevision+'"></wall>';
        try {
            var strmWall = file.open(null, opera.io.filemode.WRITE);
            strmWall.write(sWall);
            strmWall.close();
        }
        catch (e) { return false }
        return true
    };

    /**
     * Removes given entries from xml wall document;
     *
     * @private
     * @param {Array} dates array of Strings corresponding to 'date'
     * attribute of xml wall entry
     * @param {XMLDocument} xmlDoc xml document from which the data
     * is to be removed
     * @param {String} sRevision new value of revision attribute to be set
     */
    function removeDataFromXml(dates, xmlDoc, sRevision) {

        var entries =  xmlDoc.getElementsByTagName('entry');

/*
var toBeRemoved = Array.prototype.filter.call(entries, function(entry) {
    var entryDate = entry.getAttribute('date');
    dates.some( function (date) { return (date == entryDate) } );
});
*/
        var toBeRemoved = [];
        for (var i = 0, entry; entry = entries[i]; ++i) {
            var entryDate = entry.getAttribute('date');
            if (dates.some( function (date) { return (date == entryDate) } )) {
                toBeRemoved[toBeRemoved.length] = entry;
            }
        }

        for (var i = 0, removeMe; removeMe = toBeRemoved[i]; ++i) {
            removeMe.parentNode.removeChild(removeMe);
        }

        var wall = xmlDoc.getElementsByTagName('wall')[0];
        wall.setAttribute('revision',sRevision);

    };

    /**
     * sets new revision, keeps old value in prevRevision
     */
    function newRevisionValue() {
        prevRevision = revision;
        revision = (new Date()).valueOf();
    };

    /* PUBLIC */

    /**
     * Returns XMLDocument representing the service's data.
     *
     * Also reads and sets model's revision and readonly attributes.
     * @returns {XMLDocument} representing the service's data.
     */
    this.getXmlData = function() {
        xmlWall = parseXmlFromFile(fileWall)
        if (xmlWall) {
            revision
                = parseInt(xmlWall.documentElement.getAttribute('revision'));
        }
        return xmlWall
    };

    /**
     * Saves post.
     *
     * @param {Object} entryData JSON equivalent of xml entry.
     * @returns {XMLDocumentFragment} xml entry or null if the data could not
     * be saved.
     */
    this.savePost = function(entryData) {
        newRevisionValue();
        var xmlEntry = addDataToXml(entryData, xmlWall, revision);
        if (saveXmlInFile(xmlWall, fileWall)) { return xmlEntry; }

        return null;
        // there's no need to remove unsuccessful xmlEntry from xmlWall,
        // as the latter will be overwritten on the next request, anyway
    };

    /**
     * Creates new xml file and document to be used as service's database.
     *
     * @returns {Boolean} true on success, false otherwise
     */
    this.createNewXml = function() {
        newRevisionValue();
        return createNewXmlInFile(fileWall, revision)
    };

    /**
     * Gets service's current revision.
     *
     * @returns {int} service's revision.
     */
    this.getRevision = function() { return revision };

    /**
     * Gets service's previous (last) revision.
     *
     * @returns {int} service's previous revision.
     */
    this.getPrevRevision = function() { return prevRevision };

    /**
     * Removes posts from the model.
     *
     * @param {Array} aDates an array of dates indicating posts to be removed.
     */
    this.removeEntries = function(aDates) {
        newRevisionValue();
        removeDataFromXml(aDates, xmlWall,revision);
        saveXmlInFile(xmlWall, fileWall);
    };

    this.getEntriesNumber = function() {
        if (xmlWall) {
            return xmlWall.getElementsByTagName('entry').length;
        }
        return 0;
    }

};




/**
 * Constructs a new WallView object. Inteded as a Singleton.
 *
 * @class This is View part of the MVC based Fridge Notes service. Provides
 * methods building (x)html structure of the client side.
 * @constructor
 * @return A new wall view
 */
function WallView() {

    var conf = {
        templatePath : 'template.html',
        elementsTags : {    // xmlElement : htmlElement.cssClass
            entry : 'div.entry',
            author : 'div.author',
            email :    'div.email',
            post : 'div.post'
        },
        entryClasses : ['pink', 'blue', 'yellow'],
        editor : {
            FONT_MIN : 30,
            FONT_MAX : 60,
            CLIENT_HEIGHT : 169,
            POST_FACTOR : 0.44, // post's height / editor height
            STYLE : 'padding: 0px; width: 265px; position: absolute; line-height: 1.2em; font-family: Gilles_Handwriting, Verdana, sans-serif;'
        }
    };

    /**
     * Creates a complete webpage.
     *
     * Builds HTMLDocument representing a coplete page.
     * @private
     * @param {HTMLDocumentFragment} htmlFrag page's main content
     * @param {Object} usrData object info whether they're admin and/or
     * which posts they own.
     * @param {Object} pref contains info about readonly, capacity and how
     * many entries there are.
     * @param {String} templatePath a path to the template html document
     * (the pages frame).
     * @returns {HMLTDocument} representing a coplete page.
     */
    function completePage(htmlFrag, usrData, pref, templatePath) {
        var readonly = pref.readonly;

        var htmlDoc = document.implementation.createDocument("","",null);
        var shoutBox;    // HTMLElement to display wall's posts

        htmlDoc.async = false;
        htmlDoc.load(templatePath);        // load new template

        shoutBox = htmlDoc.getElementsByAttribute('id','shoutBox','div')[0];
        shoutBox.appendChild(htmlFrag);

        var form, element;

        if (!usrData.isAdmin) {
            // admin form
            form = htmlDoc.getElementsByAttribute('id','admin','div')[0];
            if (form) { form.parentNode.removeChild(form); }

            // admin script
            element = htmlDoc.getElementsByAttribute(
                'src', 'admin.js', 'script'
            )[0];
            if (element) { element.parentNode.removeChild(element); }

            // admin css
            element = htmlDoc.getElementsByAttribute(
                'href', 'admin.css', 'link'
            )[0];
            if (element) { element.parentNode.removeChild(element); }

            // readonly disabilities ;]
            if (
                readonly || (usrData.isAdmin ? false
                : pref.capacity <= pref.entriesNumber)
            ) {
                // non-js
                form = htmlDoc.getElementsByAttribute('id','shout','form')[0];

                (function(tagname) {
                    var inputs = form.getElementsByTagName(tagname);
                    for (var i = 0, input; input = inputs[i]; ++i)
                        input.setAttribute('disabled','true');
                    return arguments.callee;
                })('input')('textarea')('button');

                htmlDoc.getElementsByAttribute(
                    'id', 'send', 'button'
                )[0].setAttribute('class','disabled');

                // js
                htmlDoc.getElementsByAttribute(
                    'id','add_note_wrapper','form'
                )[0].style.display = 'none';
            }
        }
        else {
            var capacity = htmlDoc.getElementsByAttribute(
                'id','capacity','input'
            )[0];

            if (capacity) { capacity.setAttribute('value', pref.capacity); }

            if (readonly) {
                element = htmlDoc.getElementsByAttribute(
                    'id', 'readonly', 'input'
                )[0];
                element.setAttribute('checked','checked');
            }

            var servicePath = htmlDoc.getElementsByAttribute(
                'id','servicePath','input'
            )[0];

            if (servicePath) {
                servicePath.setAttribute(
                    'value',
                    "http://" + event.target.hostName 
                        + event.target.currentServicePath
                );
            }
        }

        // prefilled user's data
        if (usrData.user) {
            var keys = ['author', 'email'];
            for (var i = 0, key; key = keys[i]; ++i) {
                el = htmlDoc.getElementsByAttribute(
                    'id', key, 'input'
                )[0];

                if (el) {
                    el.setAttribute('value', usrData.user[key]);
                    if (usrData.user[key] == '') {
                        el.setAttribute('class', 'empty');
                    }
                }
            }
        
            // token
            form = htmlDoc.getElementsByAttribute('id','shout','form')[0];

            var hid_token = document.createElement('input');
            hid_token.type = 'hidden';
            hid_token.name = 'token';
            hid_token.value = usrData.user.token;
            form.lastChild.appendChild(hid_token);
        }

        return htmlDoc;
    };

    /**
     * Builds an html entry frame.
     *
     * Builds an html object to which post's text, atuhor's name etc. will be
     * appended. Creates means to remember entry's revision on the client side.
     * @private
     * @param {Object} entryData object with entry's data (post's text, author,
     * revision etc).
     * @param {Object} usrData contains info whether the user the entry is
     * built is an admin and/or what posts they own.
     * @returns {HTMLElement} representing an entry to which post's data may
     * be added.
     */
    function buildEntryHTML(entryDate, usrData) {

        var entryContent = document.createDocumentFragment();

        var dateElmnt = document.createElement('div');
        dateElmnt.className = 'date';
        dateElmnt.appendChild(document.createTextNode(
                (new Date(+entryDate)).humanReadable() + ' GMT'
            ));
        entryContent.appendChild(dateElmnt);

        var form = document.createElement('form');
        //form.name = 'removeForm';
        form.className = 'note_form';
        form.appendChild(document.createElement('div'));
        form.action = '';
        form.method = 'post';

        var hid_remove = document.createElement('input');
        hid_remove.type = 'hidden';
        hid_remove.name = 'remove';
        hid_remove.value = entryDate;
        form.lastChild.appendChild(hid_remove);

        if (
            (usrData.isAdmin) ||
            (usrData.user && ~usrData.user.entries.indexOf(entryDate))
        ) {
            var submit = document.createElement('button');
            submit.type = 'submit';
            submit.name = 'delete';
            submit.value = 'delete';
            submit.className = 'trash';
            // 4 IE
            submit.appendChild(document.createTextNode(' '));

            form.lastChild.appendChild(submit);

            form.className += ' deletable';
        }

        var hid_token = document.createElement('input');
        hid_token.type = 'hidden';
        hid_token.name = 'token';
        hid_token.value = usrData.user.token;
        form.lastChild.appendChild(hid_token);
        
        
        entryContent.appendChild(form);

        return entryContent;
    };

    /**
     * Creates html counterpart of XML wall fragment
     * @private
     * @param {XMLDocument} xmlFrag xml wall data to be transformed
     * @param {Object} usrData object used for info as to how and what
     * to transform
     * @returns {HTMLDocumentFragment} the transformation result
     */
    function xmlToHtml(xmlFrag, usrData) {

        var htmlFragment = document.createDocumentFragment();

        var xmlSerial = new XMLSerializer;
        var entries = xmlFrag.getElementsByTagName('entry');
        var elements;

        for (var i=0, entry; entry = entries[i]; ++i) {

            elements = [];
            // build elements
            for ( xmlTag in conf.elementsTags ) {

                if (!usrData.isAdmin && xmlTag == 'email') { continue }

                var arr = conf.elementsTags[xmlTag].split('.');

                var element = document.createElement(arr.shift());
                element.className = arr.join(' ');

                var entryDate = entry.getAttribute('date');

                if (xmlTag != 'entry') {
                    var xmlEl = entry.getElementsByTagName(xmlTag)[0];
                    if (xmlTag == 'author' && xmlEl.textContent == '') {
                        element.appendChild(document.createTextNode('???'));
                    }
                    else if (xmlTag == 'post') {
                        var content = xmlSerial.serializeToString(xmlEl)
                            .replace(/<post>/g,'').replace(/<\/post>/g,'')
                            .replace(/<dbsp\/>/g,' &nbsp;');
                            

                        element.innerHTML = content.linkify({attributes : 'target="_blank"'});
                        element.title = content.replace(/<br\/>/g,' ')
                                        .replace(/&lt;/g,'<');
                        element.style.fontSize = getFontSizePx(content);
                    }
                    else {
                        element.appendChild(
                            document.createTextNode(xmlEl.textContent)
                        );
                    }

                    elements.push(element);
                }
                else {    // entry
                    element.appendChild(buildEntryHTML(entryDate, usrData));
                    element.className += ' ' + conf.entryClasses[
                        +entryDate % conf.entryClasses.length
                    ];
                    elements.unshift(element);
                }
            }

            // append elements[1:last] as children of elements[0] (entry)
            for (var j = 1, el; el = elements[j]; ++j) {
                elements[0].appendChild(el);
            }
            elements[0].appendChild(element);

            // last first
            htmlFragment.insertBefore(elements[0], htmlFragment.firstChild);
        }

        return htmlFragment;
    };

    var getFontSizePx = (function() {
        var editor = conf.editor;
        var origin = {};

        origin.clientHeight = editor.CLIENT_HEIGHT;
        origin.style = editor.STYLE;

        var fontSetter = window.getFontSetter(origin, editor.FONT_MIN, editor.FONT_MAX);

        return function(string) {
            var fontSize = fontSetter.computeFont(string);
            return (fontSize * editor.POST_FACTOR >> 0) + 'px';
        }
    })();

    /**
     * Creates a complete webpage.
     *
     * Builds HTMLDocument representing a coplete page.
     * @private
     * @param {HTMLDocumentFragment} htmlFrag page's main content
     * @param {Object} usrData object info whether they're admin and/or
     * which posts they own.
     * @param {Object} pref contains info about readonly, capacity and how
     * many entries there are.
     * @returns {HMLTDocument} representing a coplete page.
     */
    this.completePage = function(htmlFrag, usrData, pref) {
        return completePage(htmlFrag, usrData, pref, conf.templatePath);
    };

    /**
     * Creates html counterpart of XML wall fragment
     * @param {XMLDocument} xmlFrag xml wall data to be transformed
     * @param {Object} usrData object used for info as to how and what
     * to transform
     * @returns {HTMLDocumentFragment} the transformation result
     */
    this.xmlToHtml = xmlToHtml;

};
